#include <turbo/turbo.h>
#include "icrypto.h"

#include <stdlib.h>
#include <string.h>

u8 crypto_key_setup (u8 * key, u8 keylen, crypto_key * skey, u8 dir)
{
#ifdef USE_AES
  if (dir == KEY_ENCRYPT)
    aes_set_encrypt_key (key, keylen, skey);
  else
    aes_set_decrypt_key (key, keylen, skey);
  return NO_ERROR;
#else
  return twofish_setup (key, keylen, skey);
#endif
}

void crypto_ecb_decrypt (u8 * ct, u8 * pt, crypto_key * key)
{
#ifdef USE_AES
  aes_decrypt (ct, pt, key);
#else
  twofish_ecb_decrypt (ct, pt, key);
#endif
}

void crypto_ecb_encrypt (u8 * ct, u8 * pt, crypto_key * key)
{
#ifdef USE_AES
  aes_encrypt (ct, pt, key);
#else
  twofish_ecb_encrypt (ct, pt, key);
#endif
}

#ifndef USE_AES_CTR
u8 crypto_cbc_decrypt (u8 * in, u8 * out, u8 nr_blocks, crypto_key * key)
{
#ifdef USE_AES
  u8 k, j;
  u8 *buf;
  u8 *inp;

  buf = malloc (2 * 16);
  if (buf == NULL)
    return ERR_NO_RAM;
  inp = buf + 16;

  for (k = 0; k < nr_blocks; k++)
  {
    crypto_ecb_decrypt (in + k * 16, buf, key);
    if (k)
      for (j = 0; j < 16; j++)
	buf[j] = buf[j] ^ inp[j];
    memcpy (inp, in + k * 16, 16);
    memcpy (out + k * 16, buf, 16);
  }

  free (buf);
  return NO_ERROR;

#else
  return twofish_cbc_decrypt (in, out, nr_blocks, key);
#endif
}

u8 crypto_cbc_encrypt (u8 * in, u8 * out, u8 nr_blocks, crypto_key * key)
{
#ifdef USE_AES
  u8 k, j;
  u8 *buf;
  u8 *tmp;

  buf = malloc (2 * 16);
  if (buf == NULL)
    return ERR_NO_RAM;
  tmp = buf + 16;

  for (k = 0; k < nr_blocks; k++)
  {
    memcpy (tmp, in + k * 16, 16);
    if (k)
      for (j = 0; j < 16; j++)
	tmp[j] ^= buf[j];
    crypto_ecb_encrypt (tmp, buf, key);
    memcpy (out + k * 16, buf, 16);
  }

  free (buf);
  return NO_ERROR;

#else
  return twofish_cbc_encrypt (in, out, nr_blocks, key);
#endif
}
#else
u8 crypto_ctr (u8 * in, u8 * out, u8 nr_blocks, u8 * ptkey, crypto_key * key,
	       u8 dir)
{
/*
Notes:
- key is already in state dir, at the end we must recreate it in case we
decrypt (we use decrypt to get counter, then setup to encrypt)

- ctr[16] is plaintext[0-7] (nonce) and 8 zeros
*/

  u8 *ctr;
  u8 *pad;
  u8 padlen;
  u8 i;
  u8 len;

  ctr = malloc (2 * 16);
  if (ctr == NULL)
    return ERR_NO_RAM;
  pad = ctr + 16;

  if (dir == KEY_DECRYPT)
  {
    dbsp ("ctr_decrypt\n");
    crypto_ecb_decrypt (in, out, key);

    for (i = 0; i < 16; i++)
      if (i < 8)
	ctr[i] = out[i];
      else
	ctr[i] = 0;

#ifdef DEBUG
    dbsp ("CTR ");
    for (i = 0; i < 16; i++)
    {
      dbch (ctr[i]);
      dbc (' ');
    }
    dbc ('\n');
#endif

    crypto_key_setup (ptkey, 16, key, KEY_ENCRYPT);
  }
  else
  {
    dbsp ("ctr_encrypt\n");

    for (i = 0; i < 16; i++)
      if (i < 8)
	ctr[i] = in[i];
      else
	ctr[i] = 0;

#ifdef DEBUG
    dbsp ("CTR ");
    for (i = 0; i < 16; i++)
    {
      dbch (ctr[i]);
      dbc (' ');
    }
    dbc ('\n');
#endif

    crypto_ecb_encrypt (in, out, key);
  }

  crypto_ecb_encrypt (ctr, pad, key);

#ifdef DEBUG
  dbsp ("PAD1 ");
  for (i = 0; i < 16; i++)
  {
    dbch (pad[i]);
    dbc (' ');
  }
  dbc ('\n');
#endif

  padlen = 0;
  len = (nr_blocks - 1) * 16;

  dbsp ("LEN: ");
  dbch (len);
  dbc ('\n');

// skip first block
  in += 16;
  out += 16;

  while (len-- > 0)
  {
    if (padlen == 16)
    {
// new ctr
      ctr[15]++;
      crypto_ecb_encrypt (ctr, pad, key);

#ifdef DEBUG
      dbsp ("PAD ");
      for (i = 0; i < 16; i++)
      {
	dbch (pad[i]);
	dbc (' ');
      }
      dbc ('\n');
#endif

      padlen = 0;
    }

    *out++ = *in++ ^ pad[padlen++];
  }

  free (ctr);

  if (dir == KEY_DECRYPT)
  {
// reset key
    crypto_key_setup (ptkey, 16, key, KEY_DECRYPT);
  }
  return NO_ERROR;
}
#endif
